#include "hyb_matrix.hpp"

PYVCL_SUBMODULE(hyb_matrix)
{
  EXPORT_HYB_MATRIX(float);
  EXPORT_HYB_MATRIX(double);
  //EXPORT_HYB_MATRIX(int);
  //EXPORT_HYB_MATRIX(long);
  //EXPORT_HYB_MATRIX(uint);
  //EXPORT_HYB_MATRIX(long);
}

