//  boost thread_clock.cpp  -----------------------------------------------------------//

//  Copyright Vicente J. Botet Escriba 2010

//  Distributed under the Boost Software License, Version 1.0.
//  See http://www.boost.org/LICENSE_1_0.txt

//  See http://www.boost.org/libs/chrono for documentation.

//--------------------------------------------------------------------------------------//

#include <boost/chrono/config.hpp>
#include <boost/chrono/detail/inlined/posix/thread_clock.hpp>
#include <cassert>

