#ifndef VIENNACL_MISC_BANDWIDTH_REDUCTION_HPP
#define VIENNACL_MISC_BANDWIDTH_REDUCTION_HPP

/* =========================================================================
   Copyright (c) 2010-2014, Institute for Microelectronics,
                            Institute for Analysis and Scientific Computing,
                            TU Wien.
   Portions of this software are copyright by UChicago Argonne, LLC.

                            -----------------
                  ViennaCL - The Vienna Computing Library
                            -----------------

   Project Head:    Karl Rupp                   rupp@iue.tuwien.ac.at

   (A list of authors and contributors can be found in the PDF manual)

   License:         MIT (X11), see file LICENSE in the base directory
============================================================================= */


/** @file viennacl/misc/bandwidth_reduction.hpp
    @brief Convenience include for bandwidth reduction algorithms such as Cuthill-McKee or Gibbs-Poole-Stockmeyer.  Experimental.
*/

#include "viennacl/misc/cuthill_mckee.hpp"
#include "viennacl/misc/gibbs_poole_stockmeyer.hpp"


namespace viennacl
{
  //TODO: Add convenience overload here. Which should be default?


} //namespace viennacl


#endif
