// Copyright (C) 2007 Joseph Gauterin
//
// Distributed under the Boost Software License, Version 1.0. (See
// accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

#ifndef BOOST_SWAP_HPP
#define BOOST_SWAP_HPP

#include "boost/utility/swap.hpp"

#endif
