#include "ell_matrix.hpp"

PYVCL_SUBMODULE(ell_matrix)
{
  EXPORT_ELL_MATRIX(float);
  EXPORT_ELL_MATRIX(double);
  //EXPORT_ELL_MATRIX(int);
  //EXPORT_ELL_MATRIX(long);
  //EXPORT_ELL_MATRIX(uint);
  //EXPORT_ELL_MATRIX(long);
}

